# known_issues.py -- Generate list of known issues for a release
#
# Richard Brooksby, Ravenbrook Limited, 2000-12-06.
#
# $Id: //info.ravenbrook.com/project/p4dti/version/1.1/tool/known_issues.py#1 $
#
#
# 1. INTRODUCTION
#
# This script generates a list of known issues for a product release,
# based on when the issues where introduced and when they were fixed.
# The information is generated from the Perforce jobs data.  Note that
# the job "Status" field is irrelevant to the list produced.
#
# See <http://info.ravenbrook.com/mail/2000/12/14/16-31-15/0.txt> and
# <http://info.ravenbrook.com/mail/2000/12/18/11-14-36/0.txt> for background.
#
# Examples of use:
#
# To list open issues in release 0.5.1 (where 6061 is the source changelist
# from which release 0.5.1 was built):
#
#     python known_issues.py -r 0.5.1 -e 6061
#
# To list issues fixed between version 0.4 and version 0.5 (where 5439 is the
# source changelist from which version 0.4 was branched and 5939 is the source
# changelist from which version 0.5 was branched):
#
#     python known_issues.py -f -s 5439 -e 5939
#
# ---
#
# This file is copyright (c) 2001 Perforce Software, Inc.  All rights  
# reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
# 
# 1.  Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
# 
# 2.  Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in
#     the documentation and/or other materials provided with the
#     distribution.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDERS AND CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
# OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
# TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
# DAMAGE.
#

import getopt
import marshal
import os
import re
import string
import sys

if os.name == 'nt' :
    mode = 'rb'
elif os.name == 'posix' :
    mode = 'r'

def p4(command):
    stream = os.popen("p4 -G " + command, mode)
    results = []
    try:
	while 1:
	    results.append(marshal.load(stream))
    except EOFError:
	status = stream.close()
    #if status:
	#raise "Perforce Error", "exit status %d" % status
    if (len(results) == 1 and
	results[0].has_key('code') and
	results[0]["code"] == "error"):
	    raise "Perforce Error", results[0]["data"]
    return results

def parse_release(string):
    match = re.match(r"\s*(\d+)\.(\d+)\.(\d+)\s*", string)
    if match:
        return map(int, match.groups())
    else:
        return None

def parse_version(string):
    match = re.match(r"\s*(\d+)\.(\d+)\s*", string)
    if match:
        return map(int, match.groups())
    else:
        return None

def print_issues_normal(issues):
    for job in issues:
        desc = string.replace(job["Description"], '\n', '\n\n')
        print "%s: %s\n\n%s" % (job["Job"], job["Title"], desc)

def print_issues_terse(issues):
    for job in issues:
        print "%s: %s" % (job["Job"], job["Title"])

def print_issues_verbose(issues):
    for job in issues:
        desc = string.replace(job["Description"], '\n', '\n\n')
        analysis = string.replace(job["Analysis"], '\n', '\n\n')
        print ("%s: %s\n\nDescription: %s\nAnalysis: %s" %
                   (job["Job"], job["Title"], desc, analysis))

priorities = ['critical', 'essential', 'optional', 'nice']

styles = { 'terse': print_issues_terse,
	   'normal': print_issues_normal,
	   'verbose': print_issues_verbose }

usage = ("Usage: python %s [options]\n"
	 "  -s,--style=     How to print the issues: terse|normal|verbose\n"
	 "  -p,--project=   The project to show issues for.\n"
	 "  --product=      The product to show issues for.\n"
	 "  -v,--version=   The version of the project (e.g., 1.2).\n"
	 "  -r,--release=   The product release (e.g., 1.2.1).\n"
	 "  -s,--start=     Start changelevel.\n"
	 "  -e,--end=       End changelevel.\n"
	 "  -f,--fixed      Show fixed issues (default is open issues).\n"
	 % sys.argv[0])

def main():
    style = 'normal'
    project = 'p4dti'
    product = None			# Defaults to project.
    version = None
    release = None
    start_changelevel = 0
    end_changelevel = int(p4('counter change')[0]['data'])
    fixed = 0

    # Parse the command line.
    try:
        opts, args = getopt.getopt(sys.argv[1:], "s:p:v:r:s:e:f",
				   ['style=', 'project=', 'product=',
				    'version=', 'release=', 'start=',
				    'end=', 'fixed'])
    except getopt.GetoptError, message:
        sys.exit(message + "\n" + usage)
    for o, a in opts:
        if o == '--style':
	    if not styles.has_key(a):
	        sys.exit("No such style: %s\n%s" % (a, usage))
            else:
                style = a
        elif o in ('-p', '--project'):
            project = a
        elif o == '--product':
            product = a
        elif o in ('-v', '--version'):
            version = parse_version(a)
            if not version:
                sys.exit("Not a valid version number.\n" + usage)
        elif o in ('-r', '--release'):
            release = parse_release(a)
            if not release:
                sys.exit("Not a valid release number.\n" + usage)
        elif o in ('-s', '--start'):
            start_changelevel = int(a)
        elif o in ('-e', '--end'):
            end_changelevel = int(a)
        elif o in ('-f', '--fixed'):
            fixed = 1
        else:
            sys.exit(usage)

    # Fill in defaults.
    if not product:
        product = project
    if version and not release:
        release = [version[0], version[1] + 1]
    elif release and not version:
        version = [release[0], release[1]]
    elif release and version:
        if version[0] != release[0] or version[1] != release[1]:
            sys.exit("Version %s doesn't match release %s\n%s"
                     % (version, release, usage))

    # Get all the issues for the project and product.
    issues = p4('jobs -e "project=%s product=%s"' % (project, product))

    # Eliminate issues which were introduced after the release we're interested
    # in.  We include any jobs which have no "Introduced-in" field, just in
    # case.
    if release:
        issues = filter(lambda job, r=release: not job.has_key("Introduced-in") or
                        parse_release(job["Introduced-in"]) <= r,
                        issues)

    # Make a filespec that gives the sources we're interested in fixes for.
    if release:
        filespec = ("//info.ravenbrook.com/project/%s/version/%d.%d/...@%d,%d"
                    % (project, version[0], version[1], start_changelevel, end_changelevel))
    else:
        filespec = ("//info.ravenbrook.com/project/%s/master/...@%d,%d"
                    % (project, start_changelevel, end_changelevel))

    # Get the list of fixed issues.
    fixes = p4("fixes -i %s" % filespec)
    fixes = filter(lambda fix: fix["Status"] == "closed", fixes)
    fixes = filter(lambda fix,s=start_changelevel:int(fix['Change'])>=s, fixes)
    fixes = filter(lambda fix,e=end_changelevel:int(fix['Change'])<=e, fixes)

    # Get the list of issues we're interested in.
    fixed_issues = map(lambda fix: fix["Job"], fixes)
    if fixed:
        issues = filter(lambda job, fi=fixed_issues: job['Job'] in fi, issues)
    else:
        issues = filter(lambda job, fi=fixed_issues: job['Job'] not in fi, issues)

    # List them by priority, severest first.
    for priority in priorities:
	print string.upper(priority)
	print
	styles[style](filter(lambda job, p=priority: job["Priority"] == p,
                             issues))

    # Report any issues that were reported as open, but are actually closed.
    if not fixed:
        closed_issues = filter(lambda job: job['Status'] == 'closed', issues)
        if closed_issues:
            print "\nTHESE ISSUES WERE REPORTED AS OPEN BUT ARE CLOSED"
            print "(This means they have no fixes.)\n"
            styles['terse'](closed_issues)

    # Check that no issue has any other priority.
    assert filter(lambda job, p=priorities: job["Priority"] not in p, issues) == []

if __name__ == '__main__':
    main()

#
# A. REFERENCES
#
#
# B. DOCUMENT HISTORY
#
# 2000-12-16 RB Created.
#
# 2000-12-18 NB Added 'master' and version tip reports.
#
# 2000-12-18 NB Added style.
#
# 2001-02-19 GDR Use command-line parameters.  Added start, fixed parameters.
#
# 2001-02-20 GDR Fixed typo in argument processing.
#
# 2001-02-23 GDR Ignore changelists outside the start/end range, to avoid
# artifacts from p4 fixes -i.
#
# 2001-03-02  RB    Transferred copyright to Perforce under their license.
#
